#! /usr/bin/env python3
"""
ft8alarm.py
(c) dg7bbp, Jens Rosebrock 2020

May be used for non commercial use for HAM radio

uses https://github.com/bmo/py-wsjtx/

with extended wsjtx_packets (send reply packets, logpacket)

"""

import pywsjtx
from pywsjtx.extra.simple_server import SimpleServer

import argparse
import threading
import re
import sys
import csv
import os
import json
import socket
import datetime
import locator

from pywsjtx.wsjtx_packets import DecodePacket, StatusPacket, QSOLoggedPacket
import subprocess

_is_windows = sys.platform == "win32"
if _is_windows:
    import winsound

_version = "1.0.1"

class QSOStatus(object):
    NO = 0
    WAIT_FOR_TRX = 1
    IN_QSO = 2
    WAIT_LOG = 3   # We are ready but don't change call until log is pressed

class AlarmServer(SimpleServer):


    def __init__(self, args, **kwargs):
        SimpleServer.__init__(self, ip_address='127.0.0.1', udp_port=args.port, **kwargs)
        self.args = args
        self.ring_on = False
        self.calls_pos = []
        self.calls_neg = []
        self.locs_pos = []
        self.locs_neg = []
        self._pause = False
        self._prepare_exit = False
        self._qso_status = QSOStatus.NO
        self._own_call = ""
        self._active_maps = set()
        self.logged_stations = set() # of tuple (call, locator, int(freq))
        self._read_list(self.calls_pos, self.calls_neg, self.args.call)
        self._read_list(self.locs_pos, self.locs_neg, self.args.locator)
        if self.args.filter:
            for l in self.args.filter:
                line = l.strip().upper()
                if line.startswith("CALL:"):
                    self._read_list(self.calls_pos, self.calls_neg, line[5:].strip().split(" "))
                elif line.startswith("LOCATOR:"):
                    self._read_list(self.locs_pos, self.locs_neg, line[8:].strip().split(" "))
        # ToDohandle file..
        if self.args.verbose > 1:
            print("Calls +: %s\n      - :%s\nLocator +: %s\n       - : %s" %
                  (self.calls_pos,
                   self.calls_neg,
                   self.locs_pos,
                   self.locs_neg))
        if not self.calls_pos and not self.calls_neg and not self.locs_pos and not self.locs_neg:
            sys.stderr.write("Need at least one filter condition\n")
            raise ValueError("No Filter")
        for r in self._read_wsjt_log():
            if self.args.verbose > 1:
                print("log: %s" % (r,))
            self.logged_stations.add(r)
        self._logbook =  hasattr(args, "logbook") and args.logbook
    
    def _read_list(self, pos_list, neg_list, inp_list):
        if inp_list:
            for l in inp_list:
                if l:
                    if l.startswith("!="):
                        neg_list.append(l[2:].upper())
                    else:
                        pos_list.append(l.upper())

    def _read_wsjt_log(self):
        fname = os.path.join(os.environ["HOME"],
                             ".local",
                             "share",
                             "WSJT-X",
                             "wsjtx.log")
        if os.path.isfile(fname):
            with open(fname,"r", newline='') as f:
                reader = csv.DictReader(f,
                    dialect="excel",
                    fieldnames=["date_from",
                                "time_from",
                                "date_to",
                                "time_to",
                                "call",
                                "locator",
                                "frequency",
                                "rstsend",
                                "rstrcvd",
                                "power"])
                for r in reader:
                    yield (r["call"], r["locator"], int(float(r["frequency"])))
        else:
            print("No wsjt/x logfile {} found.".format(fname))

    def _show_map(self, loc):
        center = loc+"ll"
        p = locator.toLoc(center)
        url = "https://openstreetmap.org/?mlat={lat}&mlon={lon}#map=2/{lat:.2}/{lon:.2}".format(lat=p.y, lon=p.x)
        subprocess.run(["chromium-browser", url])

    def show_map(self, loc):
        if args.map:
            if loc not in self._active_maps:
                self._active_maps.add(loc)
                t = threading.Thread(target=self._show_map, args=(loc,))
                t.start()

    def send_alarm(self, call, loc):
        if args.verbose > 1:
            print("Received CQ from: %s [%s]" % (call, loc))
        self._qso_status = QSOStatus.WAIT_FOR_TRX
        self.play_sound()
                
    def play_sound(self, force=False):
        """
        Plays sound in a second thread
        """
        if args.ring or force:
            if not self.ring_on:
                self.ring_on = True
                t = threading.Thread(target=self._play_sound)
                t.start()

    def _play_sound(self):
        global _is_windows
        if _is_windows:
            winsound.PlaySound(self.args.ringtone, winsound.SND_FILENAME)
        else:
            p_list = ["aplay"]
            if self.args.audiodev:
                p_list.append("-D")
                p_list.append(self.args.audiodev)
            p_list.append(self.args.ringtone)
            subprocess.run(p_list, capture_output=True)
            self.ring_on = False
        
    def send_reply(self, call, decoded_packet, addr_port):
        p_content = decoded_packet.to_reply_packet()
        t = decoded_packet.time
        snr = decoded_packet.snr
        print("Set call: %s %s %s" %  (call, snr, t.isoformat()))
        self.sock.sendto(p_content, addr_port)

    def run(self):
        self.tx_status = False
        self.dial_frequency = 0
        while not self._prepare_exit:
            (pkt, addr_port) = self.rx_packet()
            if (pkt != None):
                if pkt.startswith(b"7ft8alarm"):
                    self._process_intern(pkt, addr_port)
                    continue
                the_packet = pywsjtx.WSJTXPacketClassFactory.from_udp_packet(addr_port, pkt)
                #print("TYPE: %s" %  type(the_packet))
                if isinstance(the_packet, DecodePacket) and not self._pause:
                   splitted_msg = the_packet.message.upper().split()
                   #print (splitted_msg)
                   if len(splitted_msg) > 2:
                       call = splitted_msg[-2]
                       loc = splitted_msg[-1]
                       if splitted_msg[0] == "CQ":
                           if self.args.minsnr is not None:
                               if the_packet.snr < self.args.minsnr:
                                   if self.args.verbose == 2:
                                       print("Low snr (%s). Packet discarded: %s" % (the_packet.snr,
                                                                                     the_packet.message))
                                   continue
                           if self.call_accepted(call, loc):
                               if not self.tx_status and self._qso_status <= QSOStatus.WAIT_FOR_TRX:
                                   self.send_alarm(call, loc)
                                   self.send_reply(call, the_packet, addr_port)
                                   self.show_map(loc)
                       elif self._qso_status == QSOStatus.IN_QSO:
                           # search for end, to send alarm for log
                           if splitted_msg[0] == self.own_call:
                               if self.args.verbose > 1:
                                   print("Msg to me: %s" % the_packet.message)
                               if splitted_msg[-1].find("73") >= 0:
                                   self._qso_status = QSOStatus.WAIT_LOG
                                   if self.args.verbose > 0:
                                       print ("QSO complete: %s (Wait for log)" % splitted_msg[1])
                               if self.args.ring:
                                   self.play_sound()
                       elif self.args.allmsgs and self.call_accepted(call, loc):
                           if self.args.ring:
                               self.play_sound()
                           print("Received : %s" % the_packet.message)
                elif isinstance(the_packet, StatusPacket):
                    self.tx_status = the_packet.tx_enabled != 0
                    if self.tx_status:
                        if self._qso_status == QSOStatus.WAIT_FOR_TRX:
                            self._qso_status = QSOStatus.IN_QSO
                            if self.args.verbose > 1:
                                print ("Set QSO Status in QSO")
                    else:
                        if self._qso_status == QSOStatus.IN_QSO:
                            self._qso_status = QSOStatus.WAIT_FOR_TRX
                            if self.args.verbose > 1:
                                print ("Set QSO Status back to WAIT_FOR_TRX")
                    self.own_call = the_packet.de_call
                    self.dial_frequency = the_packet.dial_frequency
                elif isinstance(the_packet, QSOLoggedPacket):
                    if self.args.verbose > 1:
                        print("QSO Logged: %s %s %s" % (the_packet.dx_call,
                                                        the_packet.dx_grid,
                                                        int(the_packet.tx_freq/1e6)))
                        print ("Set QSO Status TO NO (after Log)")
                    self._qso_status = QSOStatus.NO
                    self.logged_stations.add((the_packet.dx_call,
                                              the_packet.dx_grid,
                                              int(the_packet.tx_freq/1e6)))
                    if self._logbook:
                        self._add7log(the_packet)

    def _add7log(self, the_packet):
        """
        writes qso to 7log
        """
        mode = the_packet.mode
        env = dict()
        env.update(os.environ)
        env["DG7BBP_LOG_QRG"] =  str(the_packet.tx_freq)
        logcall = ["/usr/bin/7log",
                   "--call",
                   the_packet.dx_call,
                   "--locator",
                   the_packet.dx_grid,
                   "--mode",
                   mode,
                   "--rst",
                   the_packet.report_recv,
                   "--report",
                   the_packet.report_send]
        if self.args.verbose > 1:
            print ("call 7log with: %s\n%s\n" % (env, logcall))
        subprocess.run(logcall, env=env)

    def call_accepted(self, call, locator):
        match = False
        if not self.calls_pos and not self.locs_pos:
            match = True
        if not match:
            if self.calls_pos:
                for c in self.calls_pos:
                    if self.args.verbose > 1:
                        print ("Try call %s %s" % (c, call))                    
                    if re.match(c, call, re.I) != None:
                        match = True
                        break
        if not match:
            if self.locs_pos:
                for l in self.locs_pos:
                    if self.args.verbose > 1:
                        print ("Try loc %s %s" % (l, locator))
                    if re.match(l, locator, re.I) != None:
                        match = True
                        break
        # filter !=
        band = int(self.dial_frequency/1e6)
        if self.args.verbose > 1:
            print("Before neg filter: call: %s match: %s band: %s locator: %s" % (call,match, band, locator))
        if match and self.calls_neg:
                for c in self.calls_pos:
                    if self.args.verbose > 1:
                        print ("Try inverted call %s %s" % (c, call))
                    if re.match(c, call, re.I) != None:
                        match = False
                        if self.args.verbose > 1:
                            print ("Match %s" % c)
                        break
        if match and self.locs_neg:
                for l in self.locs_neg:
                    if self.args.verbose > 1:
                        print ("Try inverted loc %s %s" % (l, locator))
                    if re.match(l, locator, re.I) != None:
                        match = False
                        if self.args.verbose > 1:
                            print ("Match %s" % l)
                        break
        if match:
            # Filter worked stations
            match = (call, locator, band) not in self.logged_stations
        return match
    
    def _process_intern(self, pkt, addr_port):
        """
        expects pkt with "7ft8alarm<jsonmsg>"
        json_data = {"type": <cmd|-}
                     "cmd": <command>}
        """
        json_data = json.loads(pkt.decode("utf-8")[len("7ft8alarm"):])
        if self.args.verbose > 1:
            print("received intern: %s" % json_data)
        typ = json_data.get("type")
        if typ == "cmd":
            cmd = json_data.get("cmd")
            if cmd == "pause":
                self._pause = True
            elif cmd == "continue":
                self._pause = False
            elif cmd == "exit":
                self._prepare_exit = True
            
    @classmethod
    def send_cmd(cls, cmd, udp_port):
        json_data = {"type": "cmd",
                     "cmd": cmd}
        ip_address = "127.0.0.1"
        cmd_data = ("7ft8alarm%s" % json.dumps(json_data)).encode("utf-8")
        sock = socket.socket(socket.AF_INET,  # Internet
                                 socket.SOCK_DGRAM)  # UDP
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        sock.bind((ip_address, 0))
        addr_port = ("127.0.0.1", udp_port)
        sock.sendto(cmd_data,addr_port)
        sock.close()

def parse_args():
    """
    Define and parse the arguments
    """
    desc = """Sends alarms if a call or a locator matches
              the given conditions. Transfer the CQ call
              like clicking the call to qso section.
              Calls and locators are regular expressions.
              a prefixed !=<expr> negates the expression.
              The filter file consumes lines with the
              prefix call: or locator:.
           """
 
    global _is_windows
    ap = argparse.ArgumentParser(description="DG7BBP - FT8-Alarm (%s)" % _version, epilog=desc)
    d = os.path.dirname(os.path.join("..", os.path.dirname(pywsjtx.__file__)))

    default_alarm = os.path.join(d , "alarm.wav")
    ap.add_argument("--ringtone", default=default_alarm, help="Alternativ wave file for ring tone")
    ap.add_argument("--ring", action="store_true", help="Enable ring tone")
    if not _is_windows:
        ap.add_argument("--audiodev", default="plughw:CARD=b1,DEV=0", help="Audiodevice to play ring")
        #ap.add_argument("--audiodev", help="Audiodevice to play ring tone")
    ap.add_argument("--port", type=int, default=SimpleServer.DEFAULT_UDP_PORT, help="UDP listen port default: %s" % SimpleServer.DEFAULT_UDP_PORT)
    ap.add_argument("--verbose", type=int, default=1, choices=[0, 1, 2],help="0: No output, 1(Default): Alarms, 2: Debug")
    ap.add_argument("--locator", nargs="*", help="List of locator regular expression. Use != for inverting")
    ap.add_argument("--call", nargs="*", help="List of call regular expression. Use != for inverting")
    ap.add_argument("--minsnr", type=int, help="Minimum SNR for decoded packets")
    ap.add_argument("--allmsgs", action="store_true", help="Warn on all msg from sender")
    ap.add_argument("--filter", type=argparse.FileType("r"), help="A file with filter conditions")
    ap.add_argument("--cmd", choices=["pause", "continue", "exit"], help="Send local command")
    ap.add_argument("--map", action="store_true", help="show stations in openstreetmap")
    if os.path.isfile("/usr/bin/7log"):
        ap.add_argument("--7log", dest="logbook", action="store_true", help="Log qsos to 7log" )
    args = ap.parse_args()
    return args

if __name__ == "__main__":
    args = parse_args()
    if args.cmd:
        AlarmServer.send_cmd(args.cmd, args.port)
    else:
        try:
            s = AlarmServer(args)
            if args.verbose > 0:
                print("Press crtl c for cancel")
            s.run()
        except (ValueError, KeyboardInterrupt) as e:
            pass

